import warnings
from math import ceil

import numpy as np
import openmdao.api as om
from wisdem.landbosse.model.Manager import Manager
from wisdem.landbosse.model.DefaultMasterInputDict import DefaultMasterInputDict
from wisdem.landbosse.landbosse_omdao.OpenMDAODataframeCache import OpenMDAODataframeCache
from wisdem.landbosse.landbosse_omdao.WeatherWindowCSVReader import read_weather_window

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", message="numpy.ufunc size changed")
    import pandas as pd


use_default_component_data = -1.0


class LandBOSSE(om.Group):
    def setup(self):

        # Add a tower section height variable. The default value of 30 m is for transportable tower sections.
        self.set_input_defaults("tower_section_length_m", 30.0, units="m")
        self.set_input_defaults("blade_drag_coefficient", use_default_component_data)  # Unitless
        self.set_input_defaults("blade_lever_arm", use_default_component_data, units="m")
        self.set_input_defaults("blade_install_cycle_time", use_default_component_data, units="h")
        self.set_input_defaults("blade_offload_hook_height", use_default_component_data, units="m")
        self.set_input_defaults("blade_offload_cycle_time", use_default_component_data, units="h")
        self.set_input_defaults("blade_drag_multiplier", use_default_component_data)  # Unitless

        self.set_input_defaults("turbine_spacing_rotor_diameters", 4)
        self.set_input_defaults("row_spacing_rotor_diameters", 10)
        self.set_input_defaults("commissioning_pct", 0.01)
        self.set_input_defaults("decommissioning_pct", 0.15)
        self.set_input_defaults("trench_len_to_substation_km", 50.0, units="km")
        self.set_input_defaults("interconnect_voltage_kV", 130.0, units="kV")

        self.set_input_defaults("foundation_height", 0.0, units="m")
        self.set_input_defaults("blade_mass", 8000.0, units="kg")
        self.set_input_defaults("hub_mass", 15.4e3, units="kg")
        self.set_input_defaults("nacelle_mass", 50e3, units="kg")
        self.set_input_defaults("tower_mass", 240e3, units="kg")
        self.set_input_defaults("turbine_rating_MW", 1500.0, units="kW")

        self.add_subsystem("landbosse", LandBOSSE_API(), promotes=["*"])


class LandBOSSE_API(om.ExplicitComponent):
    def setup(self):
        # Clear the cache
        OpenMDAODataframeCache._cache = {}

        self.setup_inputs()
        self.setup_outputs()
        self.setup_discrete_outputs()
        self.setup_discrete_inputs_that_are_not_dataframes()
        self.setup_discrete_inputs_that_are_dataframes()

    def setup_inputs(self):
        """
        This method sets up the inputs.
        """
        self.add_input("blade_drag_coefficient", use_default_component_data)  # Unitless
        self.add_input("blade_lever_arm", use_default_component_data, units="m")
        self.add_input("blade_install_cycle_time", use_default_component_data, units="h")
        self.add_input("blade_offload_hook_height", use_default_component_data, units="m")
        self.add_input("blade_offload_cycle_time", use_default_component_data, units="h")
        self.add_input("blade_drag_multiplier", use_default_component_data)  # Unitless

        # Even though LandBOSSE doesn't use foundation height, TowerSE does,
        # and foundation height can be used with hub height to calculate
        # tower height.

        self.add_input("foundation_height", 0.0, units="m")

        self.add_input("tower_section_length_m", 30.0, units="m")
        self.add_input("nacelle_mass", 0.0, units="kg")
        self.add_input("tower_mass", 0.0, units="kg")

        # A discrete input below, number_of_blades, gives the number of blades
        # on the rotor.
        #
        # The total mass of the rotor nacelle assembly (RNA) is the following
        # sum:
        #
        # (blade_mass * number_of_blades) + nac_mass + hub_mass

        self.add_input("blade_mass", use_default_component_data, units="kg", desc="The mass of one rotor blade.")

        self.add_input("hub_mass", use_default_component_data, units="kg", desc="Mass of the rotor hub")

        self.add_input(
            "crane_breakdown_fraction",
            val=0.0,
            desc="0 means the crane is never broken down. 1 means it is broken down every turbine.",
        )

        self.add_input("construct_duration", val=9, desc="Total project construction time (months)")
        self.add_input("hub_height_meters", val=80, units="m", desc="Hub height m")
        self.add_input("rotor_diameter_m", val=77, units="m", desc="Rotor diameter m")
        self.add_input("wind_shear_exponent", val=0.2, desc="Wind shear exponent")
        self.add_input("turbine_rating_MW", val=1.5, units="MW", desc="Turbine rating MW")
        self.add_input("fuel_cost_usd_per_gal", val=1.5, desc="Fuel cost USD/gal")

        self.add_input(
            "breakpoint_between_base_and_topping_percent", val=0.8, desc="Breakpoint between base and topping (percent)"
        )

        # Could not place units in turbine_spacing_rotor_diameters
        self.add_input("turbine_spacing_rotor_diameters", desc="Turbine spacing (times rotor diameter)", val=4)

        self.add_input("depth", units="m", desc="Foundation depth m", val=2.36)
        self.add_input("rated_thrust_N", units="N", desc="Rated Thrust (N)", val=5.89e5)

        # Can't set units
        self.add_input("bearing_pressure_n_m2", desc="Bearing Pressure (n/m2)", val=191521)

        self.add_input("gust_velocity_m_per_s", units="m/s", desc="50-year Gust Velocity (m/s)", val=59.5)
        self.add_input("road_length_adder_m", units="m", desc="Road length adder (m)", val=5000)

        # Can't set units
        self.add_input("fraction_new_roads", desc="Percent of roads that will be constructed (0.0 - 1.0)", val=0.33)

        self.add_input("road_quality", desc="Road Quality (0-1)", val=0.6)
        self.add_input("line_frequency_hz", units="Hz", desc="Line Frequency (Hz)", val=60)

        # Can't set units
        self.add_input("row_spacing_rotor_diameters", desc="Row spacing (times rotor diameter)", val=10)

        self.add_input(
            "trench_len_to_substation_km", units="km", desc="Combined Homerun Trench Length to Substation (km)", val=50
        )
        self.add_input("distance_to_interconnect_mi", units="mi", desc="Distance to interconnect (miles)", val=5)
        self.add_input("interconnect_voltage_kV", units="kV", desc="Interconnect Voltage (kV)", val=130)
        self.add_input(
            "critical_speed_non_erection_wind_delays_m_per_s",
            units="m/s",
            desc="Non-Erection Wind Delay Critical Speed (m/s)",
            val=15,
        )
        self.add_input(
            "critical_height_non_erection_wind_delays_m",
            units="m",
            desc="Non-Erection Wind Delay Critical Height (m)",
            val=10,
        )
        self.add_discrete_input("road_distributed_winnd", val=False)
        self.add_input("road_width_ft", units="ft", desc="Road width (ft)", val=20)
        self.add_input("road_thickness", desc="Road thickness (in)", val=8)
        self.add_input("crane_width", units="m", desc="Crane width (m)", val=12.2)
        self.add_input("overtime_multiplier", desc="Overtime multiplier", val=1.4)
        self.add_input("markup_contingency", desc="Markup contingency", val=0.03)
        self.add_input("markup_warranty_management", desc="Markup warranty management", val=0.0002)
        self.add_input("markup_sales_and_use_tax", desc="Markup sales and use tax", val=0)
        self.add_input("markup_overhead", desc="Markup overhead", val=0.05)
        self.add_input("markup_profit_margin", desc="Markup profit margin", val=0.05)
        self.add_input("Mass tonne", val=(1.0,), desc="", units="t")
        self.add_input(
            "development_labor_cost_usd", val=1e6, desc="The cost of labor in the development phase", units="USD"
        )
        # Disabled due to Pandas conflict right now.
        self.add_input("labor_cost_multiplier", val=1.0, desc="Labor cost multiplier")

        self.add_input("commissioning_pct", 0.01)
        self.add_input("decommissioning_pct", 0.15)

    def setup_discrete_inputs_that_are_not_dataframes(self):
        """
        This method sets up the discrete inputs that aren't dataframes.
        """
        self.add_discrete_input("num_turbines", val=100, desc="Number of turbines in project")

        # Since 3 blades are so common on rotors, that is a reasonable default
        # value that will not need to be checked during component list
        # assembly.

        self.add_discrete_input("number_of_blades", val=3, desc="Number of blades on the rotor")

        self.add_discrete_input(
            "user_defined_home_run_trench", val=0, desc="Flag for user-defined home run trench length (0 = no; 1 = yes)"
        )

        self.add_discrete_input(
            "allow_same_flag",
            val=False,
            desc="Allow same crane for base and topping (True or False)",
        )

        self.add_discrete_input(
            "hour_day",
            desc="Dictionary of normal and long hours for construction in a day in the form of {'long': 24, 'normal': 10}",
            val={"long": 24, "normal": 10},
        )

        self.add_discrete_input(
            "time_construct",
            desc="One of the keys in the hour_day dictionary to specify how many hours per day construction happens.",
            val="normal",
        )

        self.add_discrete_input(
            "user_defined_distance_to_grid_connection",
            desc="Flag for user-defined home run trench length (True or False)",
            val=False,
        )

        # Could not place units in rate_of_deliveries
        self.add_discrete_input("rate_of_deliveries", val=10, desc="Rate of deliveries (turbines per week)")

        self.add_discrete_input("new_switchyard", desc="New Switchyard (True or False)", val=True)
        self.add_discrete_input("num_hwy_permits", desc="Number of highway permits", val=10)
        self.add_discrete_input("num_access_roads", desc="Number of access roads", val=2)

    def setup_discrete_inputs_that_are_dataframes(self):
        """
        This sets up the default inputs that are dataframes. They are separate
        because they hold the project data and the way we need to hold their
        data is different. They have defaults loaded at the top of the file
        which can be overridden outside by setting the properties listed
        below.
        """
        # Read in default sheets for project data
        default_project_data = OpenMDAODataframeCache.read_all_sheets_from_xlsx("ge15_public")

        self.add_discrete_input(
            "site_facility_building_area_df",
            val=default_project_data["site_facility_building_area"],
            desc="site_facility_building_area DataFrame",
        )

        self.add_discrete_input(
            "components",
            val=default_project_data["components"],
            desc="Dataframe of components for tower, blade, nacelle",
        )

        self.add_discrete_input(
            "crane_specs", val=default_project_data["crane_specs"], desc="Dataframe of specifications of cranes"
        )

        self.add_discrete_input(
            "weather_window",
            val=read_weather_window(default_project_data["weather_window"]),
            desc="Dataframe of wind toolkit data",
        )

        self.add_discrete_input("crew", val=default_project_data["crew"], desc="Dataframe of crew configurations")

        self.add_discrete_input(
            "crew_price",
            val=default_project_data["crew_price"],
            desc="Dataframe of costs per hour for each type of worker.",
        )

        self.add_discrete_input(
            "equip", val=default_project_data["equip"], desc="Collections of equipment to perform erection operations."
        )

        self.add_discrete_input(
            "equip_price", val=default_project_data["equip_price"], desc="Prices for various type of equipment."
        )

        self.add_discrete_input("rsmeans", val=default_project_data["rsmeans"], desc="RSMeans price data")

        self.add_discrete_input(
            "cable_specs", val=default_project_data["cable_specs"], desc="cable specs for collection system"
        )

        self.add_discrete_input(
            "material_price",
            val=default_project_data["material_price"],
            desc="Prices of materials for foundations and roads",
        )

        self.add_discrete_input("project_data", val=default_project_data, desc="Dictionary of all dataframes of data")

    def setup_outputs(self):
        """
        This method sets up the continuous outputs. This is where total costs
        and installation times go.

        To see how cost totals are calculated see, the compute_total_bos_costs
        method below.
        """
        self.add_output(
            "bos_capex", 0.0, units="USD", desc="Total BOS CAPEX not including commissioning or decommissioning."
        )
        self.add_output(
            "bos_capex_kW",
            0.0,
            units="USD/kW",
            desc="Total BOS CAPEX per kW not including commissioning or decommissioning.",
        )
        self.add_output(
            "total_capex", 0.0, units="USD", desc="Total BOS CAPEX including commissioning and decommissioning."
        )
        self.add_output(
            "total_capex_kW",
            0.0,
            units="USD/kW",
            desc="Total BOS CAPEX per kW including commissioning and decommissioning.",
        )
        self.add_output("installation_capex", 0.0, units="USD", desc="Total foundation and erection installation cost.")
        self.add_output(
            "installation_capex_kW", 0.0, units="USD", desc="Total foundation and erection installation cost per kW."
        )
        self.add_output("installation_time_months", 0.0, desc="Total balance of system installation time (months).")

    def setup_discrete_outputs(self):
        """
        This method sets up discrete outputs.
        """
        self.add_discrete_output(
            "landbosse_costs_by_module_type_operation", desc="The costs by module, type and operation", val=None
        )

        self.add_discrete_output(
            "landbosse_details_by_module",
            desc="The details from the run of LandBOSSE. This includes some costs, but mostly other things",
            val=None,
        )

        self.add_discrete_output("erection_crane_choice", desc="The crane choices for erection.", val=None)

        self.add_discrete_output(
            "erection_component_name_topvbase",
            desc="List of components and whether they are a topping or base operation",
            val=None,
        )

        self.add_discrete_output(
            "erection_components", desc="List of components with their values modified from the defaults.", val=None
        )

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """
        This runs the ErectionCost module using the inputs and outputs into and
        out of this module.

        Note: inputs, discrete_inputs are not dictionaries. They do support
        [] notation. inputs is of class 'openmdao.vectors.default_vector.DefaultVector'
        discrete_inputs is of class openmdao.core.component._DictValues. Other than
        [] brackets, they do not behave like dictionaries. See the following
        documentation for details.

        http://openmdao.org/twodocs/versions/latest/_srcdocs/packages/vectors/default_vector.html
        https://mdolab.github.io/OpenAeroStruct/_modules/openmdao/core/component.html

        Parameters
        ----------
        inputs : openmdao.vectors.default_vector.DefaultVector
            A dictionary-like object with NumPy arrays that hold float
            inputs. Note that since these are NumPy arrays, they
            need indexing to pull out simple float64 values.

        outputs : openmdao.vectors.default_vector.DefaultVector
            A dictionary-like object to store outputs.

        discrete_inputs : openmdao.core.component._DictValues
            A dictionary-like with the non-numeric inputs (like
            pandas.DataFrame)

        discrete_outputs : openmdao.core.component._DictValues
            A dictionary-like for non-numeric outputs (like
            pandas.DataFrame)
        """

        # Put the inputs together and run all the modules
        master_output_dict = dict()
        master_input_dict = self.prepare_master_input_dictionary(inputs, discrete_inputs)
        manager = Manager(master_input_dict, master_output_dict)
        result = manager.execute_landbosse("WISDEM")

        # Check if everything executed correctly
        if result != 0:
            raise Exception("LandBOSSE didn't execute correctly")

        # Gather the cost and detail outputs

        costs_by_module_type_operation = self.gather_costs_from_master_output_dict(master_output_dict)
        discrete_outputs["landbosse_costs_by_module_type_operation"] = costs_by_module_type_operation

        details = self.gather_details_from_master_output_dict(master_output_dict)
        discrete_outputs["landbosse_details_by_module"] = details

        # This is where we have access to the modified components, so put those
        # in the outputs of the component
        discrete_outputs["erection_components"] = master_input_dict["components"]

        # Now get specific outputs. These have been refactored to methods that work
        # with each module so as to keep this method as compact as possible.
        self.gather_specific_erection_outputs(master_output_dict, outputs, discrete_outputs)

        # Compute the total BOS costs
        self.compute_total_bos_costs(costs_by_module_type_operation, master_output_dict, inputs, outputs)

    def prepare_master_input_dictionary(self, inputs, discrete_inputs):
        """
        This prepares a master input dictionary by applying all the necessary
        modifications to the inputs.

        Parameters
        ----------
        inputs : openmdao.vectors.default_vector.DefaultVector
            A dictionary-like object with NumPy arrays that hold float
            inputs. Note that since these are NumPy arrays, they
            need indexing to pull out simple float64 values.

        discrete_inputs : openmdao.core.component._DictValues
            A dictionary-like with the non-numeric inputs (like
            pandas.DataFrame)

        Returns
        -------
        dict
            The prepared master input to go to the Manager.
        """
        inputs_dict = {key: inputs[key][0] for key in inputs.keys()}
        discrete_inputs_dict = {key: value for key, value in discrete_inputs.items()}
        incomplete_input_dict = {**inputs_dict, **discrete_inputs_dict}

        # Modify the default component data if needed and copy it into the
        # appropriate values of the input dictionary.
        modified_components = self.modify_component_lists(inputs, discrete_inputs)
        incomplete_input_dict["project_data"]["components"] = modified_components
        incomplete_input_dict["components"] = modified_components

        # FoundationCost needs to have all the component data split into separate
        # NumPy arrays.
        incomplete_input_dict["component_data"] = modified_components
        for component in incomplete_input_dict["component_data"].keys():
            incomplete_input_dict[component] = np.array(incomplete_input_dict["component_data"][component])

        # These are aliases because parts of the code call the same thing by
        # difference names.
        incomplete_input_dict["crew_cost"] = discrete_inputs["crew_price"]
        incomplete_input_dict["cable_specs_pd"] = discrete_inputs["cable_specs"]

        # read in RSMeans per diem:
        crew_cost = discrete_inputs["crew_price"]
        crew_cost = crew_cost.set_index("Labor type ID", drop=False)
        incomplete_input_dict["rsmeans_per_diem"] = crew_cost.loc["RSMeans", "Per diem USD per day"]

        # Calculate project size in megawatts
        incomplete_input_dict["project_size_megawatts"] = float(
            discrete_inputs["num_turbines"] * inputs["turbine_rating_MW"]
        )

        # Needed to avoid distributed wind keys
        incomplete_input_dict["road_distributed_wind"] = False

        defaults = DefaultMasterInputDict()
        master_input_dict = defaults.populate_input_dict(incomplete_input_dict)

        return master_input_dict

    def gather_costs_from_master_output_dict(self, master_output_dict):
        """
        This method extract all the cost_by_module_type_operation lists for
        output in an Excel file.

        It finds values for the keys ending in '_module_type_operation'. It
        then concatenates them together so they can be easily written to
        a .csv or .xlsx

        On every row, it includes the:
            Rotor diameter m
            Turbine rating MW
            Number of turbines

        This enables easy mapping of new columns if need be. The columns have
        spaces in the names so that they can be easily written to a user-friendly
        output.

        Parameters
        ----------
        runs_dict : dict
            Values are the names of the projects. Keys are the lists of
            dictionaries that are lines for the .csv

        Returns
        -------
        list
            List of dicts to write to the .csv.
        """
        line_items = []

        # Gather the lists of costs
        cost_lists = [value for key, value in master_output_dict.items() if key.endswith("_module_type_operation")]

        # Flatten the list of lists that is the result of the gathering
        for cost_list in cost_lists:
            line_items.extend(cost_list)

        # Filter out the keys needed and rename them to meaningful values
        final_costs = []
        for line_item in line_items:
            item = {
                "Module": line_item["module"],
                "Type of cost": line_item["type_of_cost"],
                "Cost / kW": line_item["usd_per_kw_per_project"],
                "Cost / project": line_item["cost_per_project"],
                "Cost / turbine": line_item["cost_per_turbine"],
                "Number of turbines": line_item["num_turbines"],
                "Rotor diameter (m)": line_item["rotor_diameter_m"],
                "Turbine rating (MW)": line_item["turbine_rating_MW"],
                "Project ID with serial": line_item["project_id_with_serial"],
            }
            final_costs.append(item)

        return final_costs

    def gather_details_from_master_output_dict(self, master_output_dict):
        """
        This extracts the detail lists from all the modules to output
        the detailed non-cost data from the model run.

        Parameters
        ----------
        master_output_dict : dict
            The master output dict with the finished module output in it.

        Returns
        -------
        list
            List of dicts with detailed data.
        """
        line_items = []

        # Gather the lists of costs
        details_lists = [value for key, value in master_output_dict.items() if key.endswith("_csv")]

        # Flatten the list of lists
        for details_list in details_lists:
            line_items.extend(details_list)

        return line_items

    def gather_specific_erection_outputs(self, master_output_dict, outputs, discrete_outputs):
        """
        This method gathers specific outputs from the ErectionCost module and places
        them on the outputs.

        The method does not return anything. Rather, it places the outputs directly
        on the continuous of discrete outputs.

        Parameters
        ----------
        master_output_dict: dict
            The master output dictionary out of LandBOSSE

        outputs : openmdao.vectors.default_vector.DefaultVector
            A dictionary-like object to store outputs.

        discrete_outputs : openmdao.core.component._DictValues
            A dictionary-like for non-numeric outputs (like
            pandas.DataFrame)
        """
        discrete_outputs["erection_crane_choice"] = master_output_dict["crane_choice"]
        discrete_outputs["erection_component_name_topvbase"] = master_output_dict["component_name_topvbase"]

    def compute_total_bos_costs(self, costs_by_module_type_operation, master_output_dict, inputs, outputs):
        """
        This computes the total BOS costs from the master output dictionary
        and places them on the necessary outputs.

        Parameters
        ----------
        costs_by_module_type_operation: List[Dict[str, Any]]
            The lists of costs by module, type and operation.

        master_output_dict: Dict[str, Any]
            The master output dictionary from the run. Used to obtain the
            construction time,

        outputs : openmdao.vectors.default_vector.DefaultVector
            The outputs in which to place the results of the computations
        """
        bos_per_kw = 0.0
        bos_per_project = 0.0
        installation_per_project = 0.0
        installation_per_kW = 0.0

        for row in costs_by_module_type_operation:
            bos_per_kw += row["Cost / kW"]
            bos_per_project += row["Cost / project"]
            if row["Module"] in ["ErectionCost", "FoundationCost"]:
                installation_per_project += row["Cost / project"]
                installation_per_kW += row["Cost / kW"]

        commissioning_pct = inputs["commissioning_pct"]
        decommissioning_pct = inputs["decommissioning_pct"]

        commissioning_per_project = bos_per_project * commissioning_pct
        decomissioning_per_project = bos_per_project * decommissioning_pct
        commissioning_per_kW = bos_per_kw * commissioning_pct
        decomissioning_per_kW = bos_per_kw * decommissioning_pct

        outputs["total_capex_kW"] = np.round(bos_per_kw + commissioning_per_kW + decomissioning_per_kW, 0)
        outputs["total_capex"] = np.round(bos_per_project + commissioning_per_project + decomissioning_per_project, 0)
        outputs["bos_capex"] = round(bos_per_project, 0)
        outputs["bos_capex_kW"] = round(bos_per_kw, 0)
        outputs["installation_capex"] = round(installation_per_project, 0)
        outputs["installation_capex_kW"] = round(installation_per_kW, 0)

        actual_construction_months = master_output_dict["actual_construction_months"]
        outputs["installation_time_months"] = round(actual_construction_months, 0)

    def modify_component_lists(self, inputs, discrete_inputs):
        """
        This method modifies the previously loaded default component lists with
        data about blades, tower sections, if they have been provided as input
        to the component.

        It only modifies the project component data if default data for the proper
        inputs have been overridden.

        The default blade data is assumed to be the first component that begins
        with the word "Blade"

        This should take mass from the tower in WISDEM. Ideally, this should have
        an input for transportable tower 4.3, large diameter steel tower LDST 6.2m, or
        unconstrained key stone tower. Or give warnings about the boundaries
        that we assume.

        Parameters
        ----------
        inputs : openmdao.vectors.default_vector.DefaultVector
            A dictionary-like object with NumPy arrays that hold float
            inputs. Note that since these are NumPy arrays, they
            need indexing to pull out simple float64 values.

        discrete_inputs : openmdao.core.component._DictValues
            A dictionary-like with the non-numeric inputs (like
            pandas.DataFrame)

        Returns
        -------
        pd.DataFrame
            The dataframe with the modified components.
        """
        input_components = discrete_inputs["components"]

        # This list is a sequence of pd.Series instances that have the
        # specifications of each component.
        output_components_list = []

        # Need to convert kg to tonnes
        kg_per_tonne = 1000

        # Get the hub height
        hub_height_meters = inputs["hub_height_meters"][0]

        # Make the nacelle. This does not include the hub or blades.
        nacelle_mass_kg = inputs["nacelle_mass"][0]
        nacelle = input_components[input_components["Component"].str.startswith("Nacelle")].iloc[0].copy()
        if inputs["nacelle_mass"] != use_default_component_data:
            nacelle["Mass tonne"] = nacelle_mass_kg / kg_per_tonne
            nacelle["Component"] = "Nacelle"
        nacelle["Lift height m"] = hub_height_meters
        output_components_list.append(nacelle)

        # Make the hub
        hub_mass_kg = inputs["hub_mass"][0]
        hub = input_components[input_components["Component"].str.startswith("Hub")].iloc[0].copy()
        hub["Lift height m"] = hub_height_meters
        if hub_mass_kg != use_default_component_data:
            hub["Mass tonne"] = hub_mass_kg / kg_per_tonne
        output_components_list.append(hub)

        # Make blades
        blade = input_components[input_components["Component"].str.startswith("Blade")].iloc[0].copy()

        # There is always a hub height, so use that as the lift height
        blade["Lift height m"] = hub_height_meters

        if inputs["blade_drag_coefficient"][0] != use_default_component_data:
            blade["Coeff drag"] = inputs["blade_drag_coefficient"][0]

        if inputs["blade_lever_arm"][0] != use_default_component_data:
            blade["Lever arm m"] = inputs["blade_lever_arm"][0]

        if inputs["blade_install_cycle_time"][0] != use_default_component_data:
            blade["Cycle time installation hrs"] = inputs["blade_install_cycle_time"][0]

        if inputs["blade_offload_hook_height"][0] != use_default_component_data:
            blade["Offload hook height m"] = hub_height_meters

        if inputs["blade_offload_cycle_time"][0] != use_default_component_data:
            blade["Offload cycle time hrs"] = inputs["blade_offload_cycle_time"]

        if inputs["blade_drag_multiplier"][0] != use_default_component_data:
            blade["Multiplier drag rotor"] = inputs["blade_drag_multiplier"]

        if inputs["blade_mass"][0] != use_default_component_data:
            blade["Mass tonne"] = inputs["blade_mass"][0] / kg_per_tonne

        # Assume that number_of_blades always has a reasonable value. It's
        # default count when the discrete input is declared of 3 is always
        # reasonable unless overridden by another input.

        number_of_blades = discrete_inputs["number_of_blades"]
        for i in range(number_of_blades):
            component = f"Blade {i + 1}"
            blade_i = blade.copy()
            blade_i["Component"] = component
            output_components_list.append(blade_i)

        # Make tower sections
        tower_mass_tonnes = inputs["tower_mass"][0] / kg_per_tonne
        tower_height_m = hub_height_meters - inputs["foundation_height"][0]
        default_tower_section = input_components[input_components["Component"].str.startswith("Tower")].iloc[0]
        tower_sections = self.make_tower_sections(tower_mass_tonnes, tower_height_m, default_tower_section)
        output_components_list.extend(tower_sections)

        # Make the output component dataframe and return it.
        output_components = pd.DataFrame(output_components_list)
        return output_components

    @staticmethod
    def make_tower_sections(tower_mass_tonnes, tower_height_m, default_tower_section):
        """
        This makes tower sections for a transportable tower.

        Approximations:

        - Weight is distributed uniformly among the sections

        - The number of sections is either the maximum allowed by mass or
          the maximum allowed by height, to maintain transportability.

        For each tower section, calculate:
            - lift height
            - lever arm
            - surface area

        The rest of values should remain at their defaults.

        Note: Tower sections are constrained in maximum diameter to 4.5 m.
            However, their surface area is calculated with a 1.3 m radius
            to agree more closely with empirical data. Also, tower sections
            are approximated as cylinders.

        Parameters
        ----------
        tower_mass_tonnes: float
            The total tower mass in tonnes

        tower_height_m: float
            The total height of the tower in meters.

        default_tower_section: pd.Series
            There are a number of values that are kept constant in creating
            the tower sections. This series holds the values.

        Returns
        -------
        List[pd.Series]
            A list of series to be appended onto an output component list.
            It is not a dataframe, because it is faster to append to a list
            and make a dataframe once.
        """
        tower_radius = 1.3

        number_of_sections = max(ceil(tower_height_m / 30), ceil(tower_mass_tonnes / 80))

        tower_section_height_m = tower_height_m / number_of_sections

        tower_section_mass = tower_mass_tonnes / number_of_sections

        tower_section_surface_area_m2 = np.pi * tower_section_height_m * (tower_radius ** 2)

        sections = []
        for i in range(number_of_sections):
            lift_height_m = (i * tower_section_height_m) + tower_section_height_m
            lever_arm = (i * tower_section_height_m) + (0.5 * tower_section_height_m)
            name = f"Tower {i + 1}"

            section = default_tower_section.copy()
            section["Component"] = name
            section["Mass tonne"] = tower_section_mass
            section["Lift height m"] = lift_height_m
            section["Surface area sq m"] = tower_section_surface_area_m2
            section["Section height m"] = tower_section_height_m
            section["Lever arm m"] = lever_arm

            sections.append(section)

        return sections



"""This module contains common simulation logic related to vessels."""

__author__ = ["Jake Nunemaker", "Rob Hammond"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from marmot import process

from wisdem.orbit.core.defaults import process_times as pt
from wisdem.orbit.core.exceptions import ItemNotFound, MissingComponent


@process
def prep_for_site_operations(vessel, survey_required=False, **kwargs):
    """
    Performs preperation process at site.

    Parameters
    ----------
    vessel : Vessel
    depth : int | float
        Site depth (m).
    extension : int | float
        Jack-up extension length (m).

    Yields
    ------
    task_list : list
        List of tasks included in preperation process.
    """

    yield position_onsite(vessel, **kwargs)
    yield stabilize(vessel, **kwargs)

    if survey_required:
        survey_time = kwargs.get("rov_survey_time", pt["rov_survey_time"])
        yield vessel.task_wrapper(
            "RovSurvey",
            survey_time,
            constraints=vessel.transit_limits,
            **kwargs,
        )


@process
def stabilize(vessel, **kwargs):
    """
    Task representing time required to stabilize the vessel. If the vessel
    has a dynamic positioning system, this task does not take any time. If the
    vessel has a jacking system, the vessel will jackup.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Requires configured `dynamic_positioning` or
        `jacksys`.
    """

    try:
        _ = vessel.dynamic_positioning
        return

    except MissingComponent:
        pass

    try:
        jacksys = vessel.jacksys
        site_depth = kwargs.get("site_depth", 40)
        extension = kwargs.get("extension", site_depth + 10)
        jackup_time = jacksys.jacking_time(extension, site_depth)
        yield vessel.task_wrapper("Jackup", jackup_time, constraints=vessel.transit_limits, **kwargs)

    except MissingComponent:
        raise MissingComponent(vessel, ["Dynamic Positioning", "Jacking System"])


@process
def jackdown_if_required(vessel, **kwargs):
    """
    Task representing time required to jackdown the vessel, if jacking system
    is configured. If not, this task does not take anytime and the vessel is
    released.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel.
    """

    try:
        jacksys = vessel.jacksys
        site_depth = kwargs.get("site_depth", 40)
        extension = kwargs.get("extension", site_depth + 10)
        jackdown_time = jacksys.jacking_time(extension, site_depth)
        yield vessel.task_wrapper(
            "Jackdown",
            jackdown_time,
            constraints=vessel.transit_limits,
            **kwargs,
        )

    except MissingComponent:
        return


@process
def position_onsite(vessel, **kwargs):
    """
    Task representing time required to position `vessel` onsite.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Requires configured `transit_limits`.
    """

    position_time = kwargs.get("site_position_time", pt["site_position_time"])

    yield vessel.task_wrapper("Position Onsite", position_time, constraints=vessel.transit_limits)


@process
def shuttle_items_to_queue(vessel, port, queue, distance, items, **kwargs):
    """
    Shuttles a list of items from port to queue.

    Parameters
    ----------
    env : Environemt
    vessel : Vessel
    port : Port
    queue : simpy.Resource
        Queue object to shuttle to.
    distance : int | float
        Distance between port and site (km).
    items : list
        List of components stored as tuples to shuttle.
        - ('key', 'value')
    """

    transit_time = vessel.transit_time(distance)

    while True:

        if vessel.at_port:
            vessel.submit_debug_log(message=f"{vessel} is at port.")

            if not port.items:
                vessel.submit_debug_log(message="No items at port. Shutting down.")
                break

            # Get list of items
            try:
                yield get_list_of_items_from_port(vessel, port, items, **kwargs)

            except ItemNotFound:
                # If no items are at port and vessel.storage.items is empty,
                # the job is done
                if not vessel.storage.items:
                    vessel.submit_debug_log(message="Items not found. Shutting down.")
                    break

            # Transit to site
            vessel.update_trip_data()
            vessel.at_port = False
            yield vessel.task_wrapper("Transit", transit_time, constraints=vessel.transit_limits)
            yield stabilize(vessel, **kwargs)
            vessel.at_site = True

        if vessel.at_site:
            vessel.submit_debug_log(message=f"{vessel} is at site.")

            # Join queue to be active feeder at site
            with queue.request() as req:
                queue_start = vessel.env.now
                yield req

                queue_time = vessel.env.now - queue_start
                if queue_time > 0:
                    vessel.submit_action_log("Queue", queue_time, location="Site")

                queue.vessel = vessel
                active_start = vessel.env.now
                queue.activate.succeed()

                # Released by WTIV when objects are depleted
                vessel.release = vessel.env.event()
                yield vessel.release
                active_time = vessel.env.now - active_start

                vessel.submit_action_log("ActiveFeeder", active_time, location="Site")

                queue.vessel = None
                queue.activate = vessel.env.event()

            # Transit back to port
            vessel.at_site = False
            yield jackdown_if_required(vessel, **kwargs)
            yield vessel.task_wrapper("Transit", transit_time, constraints=vessel.transit_limits)
            vessel.at_port = True


@process
def get_list_of_items_from_port(vessel, port, items, **kwargs):
    """
    Retrieves multiples of 'items' from port until full.

    Parameters
    ----------
    vessel : Vessel
    port : Port
        Port simulation object to retrieve items from.
    items : list
        List of tuples representing items to get from port.
        - ('key': 'value')
    """

    with port.crane.request() as req:
        # Join queue to be active vessel at port
        queue_start = vessel.env.now
        yield req
        queue_time = vessel.env.now - queue_start
        if queue_time > 0:
            vessel.submit_action_log("Queue", queue_time)

        if port.items:
            while True:
                buffer = []
                for i in items:
                    item = port.get_item(i)
                    buffer.append(item)

                # Calculate deck space and mass of one complete turbine
                total_deck_space = sum([item.deck_space for item in buffer])
                proposed_deck_space = vessel.storage.current_deck_space + total_deck_space

                total_mass = sum([item.mass for item in buffer])
                proposed_mass = vessel.storage.current_cargo_mass + total_mass

                if vessel.storage.current_cargo_mass == 0:

                    if proposed_deck_space > vessel.storage.max_deck_space:

                        msg = f"Warning: '{vessel}' Deck Space Capacity Exceeded"
                        vessel.submit_debug_log(message=msg)

                    if proposed_mass > vessel.storage.max_cargo_mass:

                        msg = f"Warning: '{vessel}' Cargo Mass Capacity Exceeded"
                        vessel.submit_debug_log(message=msg)

                elif proposed_deck_space > vessel.storage.max_deck_space:
                    vessel.submit_debug_log(message="Full")
                    for item in buffer:
                        port.put(item)
                    break

                elif proposed_mass > vessel.storage.max_cargo_mass:
                    vessel.submit_debug_log(message="Full")
                    for item in buffer:
                        port.put(item)
                    break

                for item in buffer:
                    action, time = item.fasten(**kwargs)
                    vessel.storage.put_item(item)

                    if time > 0:
                        yield vessel.task_wrapper(
                            action,
                            time,
                            constraints=vessel.transit_limits,
                            **kwargs,
                        )

        else:
            raise ItemNotFound(items)


@process
def shuttle_items_to_queue_wait(vessel, port, queue, distance, items, per_trip, assigned, **kwargs):
    """
    Shuttles a list of items from port to queue.

    Parameters
    ----------
    env : Environemt
    vessel : Vessel
    port : Port
    queue : simpy.Resource
        Queue object to shuttle to.
    distance : int | float
        Distance between port and site (km).
    items : list
        List of components stored as tuples to shuttle.
        - ('key', 'value')
    per_trip : int
    assigned : int
    """

    transit_time = vessel.transit_time(distance)

    n = 0
    while n < assigned:

        vessel.submit_debug_log(message=f"{vessel} is at port.")

        # Get list of items
        per_trip = max([per_trip, 1])
        yield get_list_of_items_from_port_wait(vessel, port, items * per_trip, **kwargs)

        # Transit to site
        vessel.update_trip_data()
        yield vessel.task("Transit", transit_time, constraints=vessel.transit_limits)
        yield stabilize(vessel, **kwargs)

        vessel.submit_debug_log(message=f"{vessel} is at site.")

        # Join queue to be active feeder at site
        with queue.request() as req:
            queue_start = vessel.env.now
            yield req

            queue_time = vessel.env.now - queue_start
            if queue_time > 0:
                vessel.submit_action_log("Queue", queue_time, location="Site")

            queue.vessel = vessel
            active_start = vessel.env.now
            queue.activate.succeed()

            # Released by WTIV when objects are depleted
            vessel.release = vessel.env.event()
            yield vessel.release
            active_time = vessel.env.now - active_start

            vessel.submit_action_log("ActiveFeeder", active_time, location="Site")

            queue.vessel = None
            queue.activate = vessel.env.event()

            # Transit back to port
            vessel.at_site = False
            yield jackdown_if_required(vessel, **kwargs)
            yield vessel.task("Transit", transit_time, constraints=vessel.transit_limits)

        n += per_trip


@process
def get_list_of_items_from_port_wait(vessel, port, items, **kwargs):
    """
    Retrieves multiples of 'items' from port until full.

    Parameters
    ----------
    vessel : Vessel
    port : Port
    items : list
        List of tuples representing items to get from port.
        - ('key': 'value')
    port : Port
        Port object to get items from.
    """

    with port.crane.request() as req:
        # Join queue to be active vessel at port
        queue_start = vessel.env.now
        yield req
        queue_time = vessel.env.now - queue_start
        if queue_time > 0:
            vessel.submit_action_log("Queue", queue_time)

        for i in items:
            wait_start = vessel.env.now
            item = yield port.get(lambda x: x.type == i)
            wait_time = vessel.env.now - wait_start

            if wait_time > 0:
                vessel.submit_action_log(f"Wait for {item}", wait_time)

            action, time = item.fasten(**kwargs)
            vessel.storage.put_item(item)

            if time > 0:
                yield vessel.task(action, time, constraints=vessel.transit_limits, **kwargs)



"""
OpenMDAO Wrapper for the NLopt package of optimizers.

More info at https://nlopt.readthedocs.io/
"""


import numpy as np
import openmdao
import openmdao.utils.coloring as coloring_mod
from openmdao.core.driver import Driver, RecordingDebugging
from openmdao.utils.general_utils import simple_warning

try:
    from openmdao.utils.class_util import weak_method_wrapper as weak_method_wrapper
except ImportError:
    from openmdao.utils.class_util import WeakMethodWrapper as weak_method_wrapper


try:
    import nlopt
except ImportError:
    nlopt = None


# All optimizers in NLopt that we support and their corresponding package name.
# Other optimizers could be added, but we've focused on those that can
# handle either inequality or equality constraints.
if nlopt is not None:
    optimizer_methods = {
        "GN_DIRECT": nlopt.GN_DIRECT,
        "GN_DIRECT_L": nlopt.GN_DIRECT_L,
        "GN_DIRECT_L_NOSCAL": nlopt.GN_DIRECT_L_NOSCAL,
        "GN_ORIG_DIRECT": nlopt.GN_ORIG_DIRECT,
        "GN_ORIG_DIRECT_L": nlopt.GN_ORIG_DIRECT_L,
        "GN_AGS": nlopt.GN_AGS,
        "GN_ISRES": nlopt.GN_ISRES,
        "LN_COBYLA": nlopt.LN_COBYLA,
        "LD_MMA": nlopt.LD_MMA,
        "LD_CCSAQ": nlopt.LD_CCSAQ,
        "LD_SLSQP": nlopt.LD_SLSQP,
    }
else:
    optimizer_methods = {}

_optimizers = set(optimizer_methods)

# Define subsets of optimizers that support different functions
_gradient_optimizers = {"LD_MMA", "LD_SLSQP", "LD_CCSAQ"}
_bounds_optimizers = _optimizers
_constraint_optimizers = {
    "LD_SLSQP",
    "LN_COBYLA",
    "LD_MMA",
    "LD_CCSAQ",
    "GN_ORIG_DIRECT",
    "GN_ORIG_DIRECT_L",
    "GN_AGS",
    "GN_ISRES",
}
_constraint_grad_optimizers = _gradient_optimizers & _constraint_optimizers
_eq_constraint_optimizers = {"LD_SLSQP", "LN_COBYLA", "GN_ISRES"}
_global_optimizers = {
    "GN_DIRECT",
    "GN_DIRECT_L",
    "GN_ORIG_DIRECT",
    "GN_ORIG_DIRECT_L",
    "GN_DIRECT_L_NOSCAL",
    "GN_AGS",
    "GN_ISRES",
}

CITATIONS = """
@article{johnson_nlopt
 author = {Johnson, Steven G.},
 title = "{The NLopt nonlinear-optimization package}",
 url = {http://github.com/stevengj/nlopt},
 }
"""


class NLoptDriver(Driver):
    """
    Driver wrapper for NLopt.

    NLopt is an open-source nonlinear optimization framework.

    Attributes
    ----------
    fail : bool
        Flag that indicates failure of most recent optimization.
    iter_count : int
        Counter for function evaluations.
    result : OptimizeResult
        Result returned from NLopt.optimize call.
    _con_cache : dict
        Cached result of constraint evaluations because NLopt asks for them in a separate function.
    _con_idx : dict
        Used for constraint bookkeeping in the presence of 2-sided constraints.
    _grad_cache : OrderedDict
        Cached result of nonlinear constraint derivatives because NLopt asks for them in a separate
        function.
    _exc_info : 3 item tuple
        Storage for exception and traceback information.
    _obj_and_nlcons : list
        List of objective + nonlinear constraints. Used to compute total derivatives
        for all except linear constraints.
    _dvlist : list
        Copy of _designvars.
    _lincongrad_cache : np.ndarray
        Pre-calculated gradients of linear constraints.
    """

    def __init__(self, **kwargs):
        """
        Initialize the NLoptDriver.

        Parameters
        ----------
        **kwargs : dict of keyword arguments
            Keyword arguments that will be mapped into the Driver options.
        """
        if nlopt is None:
            raise RuntimeError("NLoptDriver is not available, NLopt is not installed.")

        super(NLoptDriver, self).__init__(**kwargs)

        # What we support
        self.supports["inequality_constraints"] = True
        self.supports["equality_constraints"] = True
        self.supports["two_sided_constraints"] = True
        self.supports["linear_constraints"] = True
        self.supports["simultaneous_derivatives"] = True

        # What we don't support
        self.supports["multiple_objectives"] = False
        self.supports["active_set"] = False
        self.supports["integer_design_vars"] = False
        self.supports._read_only = True

        self.result = None
        self._grad_cache = None
        self._con_cache = None
        self._con_idx = {}
        self._obj_and_nlcons = None
        self._dvlist = None
        self._lincongrad_cache = None
        self.iter_count = 0
        self._exc_info = None

        self.cite = CITATIONS

    def _declare_options(self):
        """
        Declare options before kwargs are processed in the init method.
        """
        self.options.declare("optimizer", "LD_SLSQP", values=_optimizers, desc="Name of optimizer to use")
        self.options.declare(
            "tol",
            1.0e-6,
            lower=0.0,
            desc="Tolerance for termination. Based on "
            + "relative function value change. Uses the "
            + "method `set_ftol_rel()` from NLOpt.",
        )
        self.options.declare("maxiter", 200, lower=0, desc="Maximum number of iterations.")
        self.options.declare("numgen", 10000, lower=0, desc="Maximum number of iterations.")
        self.options.declare(
            "maxtime",
            1.0e10,
            lower=0.0,
            desc="Maximum time in seconds to perform optimization.",
        )
        self.options.declare(
            "xtol",
            1.0e-12,
            lower=0.0,
            desc="Tolerance for termination. Based on "
            + "relative function value change. Uses the "
            + "method `set_xtol_rel()` from NLOpt.",
        )

    def _get_name(self):
        """
        Get name of current optimizer.

        Returns
        -------
        str
            The name of the current optimizer.
        """
        return "NLopt_" + self.options["optimizer"]

    def _setup_driver(self, problem):
        """
        Prepare the driver for execution.

        This is the final thing to run during setup.

        Parameters
        ----------
        problem : <Problem>
            Pointer
        """
        super(NLoptDriver, self)._setup_driver(problem)
        opt = self.options["optimizer"]

        self.supports._read_only = False
        self.supports["gradients"] = opt in _gradient_optimizers
        self.supports["inequality_constraints"] = opt in _constraint_optimizers
        self.supports["two_sided_constraints"] = opt in _constraint_optimizers
        self.supports["equality_constraints"] = opt in _eq_constraint_optimizers
        self.supports._read_only = True

        # Raises error if multiple objectives are not supported, but more objectives were defined.
        if not self.supports["multiple_objectives"] and len(self._objs) > 1:
            msg = "{} currently does not support multiple objectives."
            raise RuntimeError(msg.format(self.msginfo))

    def run(self):
        """
        Optimize the problem using selected NLopt optimizer.
        """
        problem = self._problem()
        opt = self.options["optimizer"]
        model = problem.model
        self.iter_count = 0
        self._total_jac = None

        self._check_for_missing_objective()

        # Initial Run
        with RecordingDebugging(self._get_name(), self.iter_count, self) as rec:
            model.run_solve_nonlinear()
            self.iter_count += 1

        self._con_cache = self.get_constraint_values()
        desvar_vals = self.get_design_var_values()
        self._dvlist = list(self._designvars)

        # Size Problem
        nparam = 0
        for param in self._designvars.values():
            nparam += param["size"]
        x_init = np.empty(nparam)

        # Initialize the NLopt problem with the method and number of design vars
        opt_prob = nlopt.opt(optimizer_methods[opt], int(nparam))

        # Initial Design Vars
        i = 0
        use_bounds = opt in _bounds_optimizers
        if use_bounds:
            bounds = []
        else:
            bounds = None

        # Loop through all OpenMDAO design variables and process their bounds
        for name, meta in self._designvars.items():
            size = meta["size"]
            x_init[i : i + size] = desvar_vals[name]
            i += size

            # Bounds if our optimizer supports them
            if use_bounds:
                meta_low = meta["lower"]
                meta_high = meta["upper"]
                for j in range(size):

                    if isinstance(meta_low, np.ndarray):
                        p_low = meta_low[j]
                    else:
                        p_low = meta_low

                    if isinstance(meta_high, np.ndarray):
                        p_high = meta_high[j]
                    else:
                        p_high = meta_high

                    bounds.append((p_low, p_high))

        # Actually add the bounds to the optimization problem.
        if bounds is not None:
            lower_bound, upper_bound = zip(*bounds)
            lower = np.array([x if x is not None else -np.inf for x in lower_bound])
            upper = np.array([x if x is not None else np.inf for x in upper_bound])
            opt_prob.set_lower_bounds(lower)
            opt_prob.set_upper_bounds(upper)

        # Constraints
        i = 1  # start at 1 since row 0 is the objective.  Constraints start at row 1.
        lin_i = 0  # counter for linear constraint jacobian
        lincons = []  # list of linear constraints
        self._obj_and_nlcons = list(self._objs)

        # Process and add constraints to the optimization problem.
        if opt in _constraint_optimizers:
            for name, meta in self._cons.items():
                size = meta["global_size"] if meta["distributed"] else meta["size"]
                upper = meta["upper"]
                lower = meta["lower"]
                equals = meta["equals"]

                if opt in _gradient_optimizers and "linear" in meta and meta["linear"]:
                    lincons.append(name)
                    self._con_idx[name] = lin_i
                    lin_i += size
                else:
                    self._obj_and_nlcons.append(name)
                    self._con_idx[name] = i
                    i += size

                # Loop over every index separately,
                # because it's easier to defined each
                # constraint by index.
                for j in range(size):

                    # Equality constraints are added as two inequality constraints
                    if equals is not None:
                        args = [name, False, j]
                        try:
                            opt_prob.add_equality_constraint(
                                signature_extender(weak_method_wrapper(self, "_confunc"), args)
                            )
                        except ValueError:
                            msg = (
                                "The selected optimizer, {}, does not support"
                                + " equality constraints. Select from {}."
                            )
                            raise NotImplementedError(msg.format(opt, _eq_constraint_optimizers))

                    else:
                        # Double-sided constraints are accepted by the algorithm
                        args = [name, False, j]
                        opt_prob.add_inequality_constraint(
                            signature_extender(weak_method_wrapper(self, "_confunc"), args)
                        )

                        if isinstance(upper, np.ndarray):
                            upper = upper[j]

                        if isinstance(lower, np.ndarray):
                            lower = lower[j]

                        dblcon = (upper < openmdao.INF_BOUND) and (lower > -openmdao.INF_BOUND)

                        # Add extra constraint if double-sided
                        if dblcon:
                            args = [name, True, j]
                            opt_prob.add_inequality_constraint(
                                signature_extender(weak_method_wrapper(self, "_confunc"), args)
                            )

            # precalculate gradients of linear constraints
            if lincons:
                self._lincongrad_cache = self._compute_totals(of=lincons, wrt=self._dvlist, return_format="array")
            else:
                self._lincongrad_cache = None

        # compute dynamic simul deriv coloring if option is set
        if coloring_mod._use_total_sparsity:
            if self._coloring_info["coloring"] is None and self._coloring_info["dynamic"]:
                coloring_mod.dynamic_total_coloring(self, run_model=False, fname=self._get_total_coloring_fname())

                # if the improvement wasn't large enough, turn coloring off
                info = self._coloring_info
                if info["coloring"] is not None:
                    pct = info["coloring"]._solves_info()[-1]
                    if info["min_improve_pct"] > pct:
                        info["coloring"] = info["static"] = None
                        simple_warning(
                            "%s: Coloring was deactivated.  Improvement of %.1f%% was "
                            "less than min allowed (%.1f%%)." % (self.msginfo, pct, info["min_improve_pct"])
                        )

        # Finalize the optimization problem setup and actually perform optimization
        try:
            if opt in _optimizers:
                opt_prob.set_min_objective(self._objfunc)
                opt_prob.set_ftol_rel(self.options["tol"])
                opt_prob.set_xtol_rel(self.options["xtol"])
                opt_prob.set_maxeval(int(self.options["maxiter"]))
                opt_prob.set_maxtime(self.options["maxtime"])
                opt_prob.set_population(int(self.options["maxiter"] / self.options["numgen"]))
                opt_prob.optimize(x_init)
                self.result = opt_prob.last_optimize_result()

            else:
                msg = 'Optimizer "{}" is not implemented yet. Choose from: {}'
                raise NotImplementedError(msg.format(opt, _optimizers))

        # If an exception was swallowed in one of our callbacks, we want to raise it
        except Exception as msg:
            if self._exc_info is not None:
                self._reraise()
            else:
                raise

        if self._exc_info is not None:
            self._reraise()

    def _objfunc(self, x_new, grad):
        """
        Evaluate and return the objective function.

        Model is executed here.

        Parameters
        ----------
        x_new : ndarray
            Array containing parameter values at new design point.
        grad : ndarray
            Empty array that is modified in-place with gradient information for
            the new design point.

        Returns
        -------
        float
            Value of the objective function evaluated at the new design point.
        """
        model = self._problem().model
        f_new = 1e10

        try:

            # Pass in new parameters
            i = 0
            for name, meta in self._designvars.items():
                size = meta["size"]
                self.set_design_var(name, x_new[i : i + size])
                i += size

            with RecordingDebugging(self._get_name(), self.iter_count, self) as rec:
                self.iter_count += 1

                # This is the actual model evaluation for OpenMDAO
                model.run_solve_nonlinear()

            # Get the objective function evaluations
            f_new = list(self.get_objective_values().values())[0]

            self._con_cache = self.get_constraint_values()

        except Exception as msg:
            self._exc_info = msg

        try:
            if grad.size > 0:
                self._grad_cache = self._compute_totals(
                    of=self._obj_and_nlcons, wrt=self._dvlist, return_format="array"
                )
                grad[:] = self._grad_cache[0, :]

        except Exception as msg:
            self._exc_info = msg

        return float(f_new)

    def _confunc(self, x_new, grad, name, dbl, idx):
        """
        Return the value of the constraint function requested in args.

        Note that this function is called for each constraint, so the model is only run when the
        objective is evaluated.

        Parameters
        ----------
        x_new : ndarray
            Array containing parameter values at new design point.
        grad : ndarray
            Empty array that is modified in-place with gradient information for
            the new design point.
        name : string
            Name of the constraint to be evaluated.
        dbl : bool
            True if double sided constraint.
        idx : float
            Contains index into the constraint array.

        Returns
        -------
        float
            Value of the constraint function.
        """
        if self._exc_info is not None:
            self._reraise()

        cons = self._con_cache
        meta = self._cons[name]

        if meta["linear"]:
            grad_cache = self._lincongrad_cache
        else:
            grad_cache = self._grad_cache

        grad_idx = self._con_idx[name] + idx

        # Equality constraints
        equals = meta["equals"]
        if equals is not None:
            if isinstance(equals, np.ndarray):
                equals = equals[idx]
            if grad.size > 0:
                grad[:] = grad_cache[grad_idx, :]
            return cons[name][idx] - equals

        # Note, NLopt defines constraints to be satisfied when negative,
        # which is the same as OpenMDAO.
        upper = meta["upper"]
        if isinstance(upper, np.ndarray):
            upper = upper[idx]

        lower = meta["lower"]
        if isinstance(lower, np.ndarray):
            lower = lower[idx]

        if dbl or (lower <= -openmdao.INF_BOUND):
            if grad.size > 0:
                grad[:] = grad_cache[grad_idx, :]
            return cons[name][idx] - upper
        else:
            if grad.size > 0:
                grad[:] = -grad_cache[grad_idx, :]
            return lower - cons[name][idx]

    def _reraise(self):
        """
        Reraise any exception encountered when NLopt calls back into our method.
        """
        exc = self._exc_info
        raise exc


def signature_extender(fcn, extra_args):
    """
    Closure function, which appends extra arguments to the original function call.

    The first argument is the design vector and the second is the gradient vector.
    The possible extra arguments from the callback
    of :func:`NLopt.optimize` are not passed to the function.

    Parameters
    ----------
    fcn : callable
        Function, which takes the design vector as the first argument.
    extra_args : tuple or list
        Extra arguments for the function

    Returns
    -------
    callable
        The function with the signature expected by the driver.
    """

    def closure(x, grad, *args):
        return fcn(x, grad, *extra_args)

    return closure



import numpy as np


def is_floating(prob):
    return prob.model.options["modeling_options"]["flags"]["floating"]


def is_monopile(prob):
    return prob.model.options["modeling_options"]["flags"]["monopile"]


def get_tower_diameter(prob):
    if is_monopile(prob):
        return np.r_[prob["fixedse.monopile_outer_diameter"], prob["towerse.tower_outer_diameter"][1:]]
    else:
        return prob["towerse.tower_outer_diameter"]


def get_tower_thickness(prob):
    if is_monopile(prob):
        return np.r_[prob["fixedse.monopile_wall_thickness"], prob["towerse.tower_wall_thickness"]]
    else:
        return prob["towerse.tower_wall_thickness"]


def get_zpts(prob):
    if is_monopile(prob):
        return np.r_[prob["fixedse.z_param"], prob["towerse.z_param"][1:]]
    else:
        return prob["towerse.z_param"]


def get_section_height(prob):
    return np.diff(get_zpts(prob))


def get_transition_height(prob):
    return prob["towerse.foundation_height"]


def get_tower_E(prob):
    if is_monopile(prob):
        return np.r_[prob["fixedse.member.E"], prob["towerse.member.E"][1:]]
    else:
        return prob["towerse.member.E"]


def get_tower_G(prob):
    if is_monopile(prob):
        return np.r_[prob["fixedse.member.G"], prob["towerse.member.G"][1:]]
    else:
        return prob["towerse.member.G"]


def get_tower_rho(prob):
    if is_monopile(prob):
        return np.r_[prob["fixedse.member.rho"], prob["towerse.member.rho"][1:]]
    else:
        return prob["towerse.member.rho"]


def get_tower_mass(prob):
    return prob["towerse.tower_mass"]


def get_tower_cost(prob):
    return prob["towerse.tower_cost"]


def get_monopile_mass(prob):
    return prob["fixedse.monopile_mass"]


def get_monopile_cost(prob):
    return prob["fixedse.monopile_cost"]


def get_structural_mass(prob):
    if is_monopile(prob):
        return prob["fixedse.structural_mass"]
    else:
        return prob["towerse.tower_mass"]


def get_structural_cost(prob):
    if is_monopile(prob):
        return prob["fixedse.structural_cost"]
    else:
        return prob["towerse.tower_cost"]


def get_tower_freqs(prob):
    if is_monopile(prob):
        return np.r_[prob["fixedse.structural_frequencies"], prob["towerse.tower.structural_frequencies"]]
    else:
        return prob["towerse.tower.structural_frequencies"]


def get_tower_cm(prob):
    return prob["towerse.tower_center_of_mass"]


def get_tower_cg(prob):
    return get_tower_cm(prob)



import numpy as np
import openmdao.api as om

import wisdem.commonse.utilities as util
from wisdem.commonse import eps, gravity
from wisdem.commonse.cylinder_member import NULL, MEMMAX


class FloatingConstraints(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        opt = self.options["modeling_options"]
        n_dlc = opt["WISDEM"]["n_dlc"]
        n_member = opt["floating"]["members"]["n_members"]

        self.add_input("Hsig_wave", 0.0, units="m")
        self.add_input("variable_ballast_mass", 0.0, units="kg")
        self.add_input("fairlead_radius", 0.0, units="m")
        self.add_input("fairlead", 0.0, units="m")
        self.add_input("survival_heel", 0.0, units="rad")
        tot_ball = 0
        for k in range(n_member):
            n_ball = opt["floating"]["members"]["n_ballasts"][k]
            self.add_input(f"member{k}:nodes_xyz", NULL * np.ones((MEMMAX, 3)), units="m")
            self.add_input(f"member{k}:constr_ballast_capacity", np.zeros(n_ball))
            tot_ball += n_ball
        self.add_input("platform_Iwater", 0.0, units="m**4")
        self.add_input("platform_displacement", 0.0, units="m**3")
        self.add_input("platform_center_of_buoyancy", np.zeros(3), units="m")
        self.add_input("system_center_of_mass", np.zeros(3), units="m")
        self.add_input("transition_node", np.zeros(3), units="m")

        self.add_input("turbine_F", np.zeros((3, n_dlc)), units="N")
        self.add_input("turbine_M", np.zeros((3, n_dlc)), units="N*m")
        self.add_input("max_surge_restoring_force", 0.0, units="N")
        self.add_input("operational_heel_restoring_force", np.zeros(6), units="N")
        self.add_input("survival_heel_restoring_force", np.zeros(6), units="N")

        self.add_output("constr_freeboard_heel_margin", np.zeros(n_member))
        self.add_output("constr_draft_heel_margin", np.zeros(n_member))
        self.add_output("constr_fixed_margin", np.zeros(tot_ball))
        self.add_output("constr_fairlead_wave", 0.0)
        self.add_output("constr_mooring_surge", 0.0)
        self.add_output("constr_mooring_heel", 0.0)
        self.add_output("metacentric_height", 0.0, units="m")

    def compute(self, inputs, outputs):
        opt = self.options["modeling_options"]
        n_member = opt["floating"]["members"]["n_members"]

        # Unpack inputs
        Hsig = float(inputs["Hsig_wave"])
        fairlead = np.abs(float(inputs["fairlead"]))
        R_fairlead = float(inputs["fairlead_radius"])
        max_heel = float(inputs["survival_heel"])
        cg = inputs["system_center_of_mass"]
        gamma = 1.1

        # Make sure there is sufficient margin on offset members
        # where freeboard does not get submerged and keel does not come out water
        freeboard_margin = np.zeros(n_member)
        draft_margin = np.zeros(n_member)
        ballast_margin = []
        for k in range(n_member):
            ballast_margin.extend(inputs[f"member{k}:constr_ballast_capacity"].tolist())

            xyz = inputs[f"member{k}:nodes_xyz"]
            inodes = np.where(xyz[:, 0] == NULL)[0][0]
            xyz = xyz[:inodes, :]
            xyz1 = xyz[0, :]
            xyz2 = xyz[-1, :]

            # Get xp-zp coplanar coordinates relative to cg
            xp1 = np.sqrt(np.sum((xyz1[:2] - cg[:2]) ** 2))
            zp1 = xyz1[2] - cg[2]
            xp2 = np.sqrt(np.sum((xyz2[:2] - cg[:2]) ** 2))
            zp2 = xyz2[2] - cg[2]

            # Coordinate transformation about CG and change in z-position
            xp1_h, zp1_h = util.rotate(0.0, 0.0, xp1, zp1, max_heel)
            xp2_h, zp2_h = util.rotate(0.0, 0.0, xp2, zp2, max_heel)
            dz1 = np.abs((zp1 - zp1_h) / xyz1[2])
            dz2 = np.abs((zp2 - zp2_h) / xyz2[2])

            # See if change in z-coordinate is bigger than freeboard or draft, should be <1
            if xyz1[2] > 0.0 and xyz1[2] > xyz2[2]:
                freeboard_margin[k] = dz1
            elif xyz2[2] > 0.0 and xyz2[2] > xyz1[2]:
                freeboard_margin[k] = dz2

            if xyz1[2] < 0.0 and xyz1[2] < xyz2[2]:
                draft_margin[k] = dz1
            elif xyz2[2] < 0.0 and xyz2[2] < xyz1[2]:
                draft_margin[k] = dz2

        # Ensure members have enough clearance from the waterline
        outputs["constr_freeboard_heel_margin"] = freeboard_margin
        outputs["constr_draft_heel_margin"] = draft_margin
        outputs["constr_fixed_margin"] = np.array(ballast_margin)

        # Make sure the fairlead depth is greater than the wave height with margin.  Should be <1
        outputs["constr_fairlead_wave"] = Hsig * gamma / fairlead

        # Compute the distance from the center of buoyancy to the metacentre (BM is naval architecture)
        # BM = Iw / V where V is the displacement volume (just computed)
        # Iw is the area moment of inertia (meters^4) of the water-plane cross section about the heel axis
        # For a spar, we assume this is just the I of a circle about x or y
        # See https://en.wikipedia.org/wiki/Metacentric_height
        # https://en.wikipedia.org/wiki/List_of_second_moments_of_area
        # and http://farside.ph.utexas.edu/teaching/336L/Fluidhtml/node30.html
        # Measure static stability:
        # 1. Center of buoyancy should be above CG (difference should be positive)
        # 2. Metacentric height should be positive
        Iwater_total = inputs["platform_Iwater"]
        V_platform = inputs["platform_displacement"]
        z_cb = inputs["platform_center_of_buoyancy"][2]
        buoyancy2metacentre_BM = Iwater_total / V_platform
        outputs["metacentric_height"] = buoyancy2metacentre_BM - (cg[2] - z_cb)

        # Mooring strength checks
        F_turb = inputs["turbine_F"].max(axis=1)
        M_turb = inputs["turbine_M"].max(axis=1)
        surge_restore = inputs["max_surge_restoring_force"]
        outputs["constr_mooring_surge"] = surge_restore - F_turb[0]
        heel_restore = inputs["operational_heel_restoring_force"]
        # (fairlead is assumed negative, made positive above, would otherwise be cg-fairlead)
        M_heel_restore = R_fairlead * heel_restore[2] + (cg[2] + fairlead) * heel_restore[:2].sum() + heel_restore[4]
        trans2cg = inputs["transition_node"][2] - cg[2]
        outputs["constr_mooring_heel"] = M_heel_restore - F_turb[0] * trans2cg - M_turb[1]


class RigidModes(om.ExplicitComponent):
    def setup(self):
        self.add_input("platform_Awater", 0.0, units="m**2")
        self.add_input("platform_added_mass", np.zeros(6), units="kg")
        self.add_input("system_mass", 0.0, units="kg")
        self.add_input("system_center_of_mass", np.zeros(3), units="m")
        self.add_input("system_I", np.zeros(6), units="kg*m**2")
        self.add_input("rho_water", 0.0, units="kg/m**3")
        self.add_input("platform_displacement", 0.0, units="m**3")
        self.add_input("metacentric_height", 0.0, units="m")
        self.add_input("mooring_stiffness", np.zeros((6, 6)), units="N/m")

        self.add_output(
            "hydrostatic_stiffness", np.zeros(6), units="N/m", desc="Summary hydrostatic stiffness of structure"
        )
        self.add_output("rigid_body_periods", np.zeros(6), units="s", desc="Natural periods of oscillation in 6 DOF")
        self.add_output("surge_period", 0.0, units="s", desc="Surge period of oscillation")
        self.add_output("sway_period", 0.0, units="s", desc="Sway period of oscillation")
        self.add_output("heave_period", 0.0, units="s", desc="Heave period of oscillation")
        self.add_output("roll_period", 0.0, units="s", desc="Roll period of oscillation")
        self.add_output("pitch_period", 0.0, units="s", desc="Pitch period of oscillation")
        self.add_output("yaw_period", 0.0, units="s", desc="Yaw period of oscillation")

    def compute(self, inputs, outputs):
        # Number of degrees of freedom
        nDOF = 6

        # Unpack variables
        rhoWater = inputs["rho_water"]
        m_system = inputs["system_mass"]
        I_system = inputs["system_I"]
        z_cg = inputs["system_center_of_mass"][-1]
        A_mat = inputs["platform_added_mass"]
        V_system = inputs["platform_displacement"]
        h_metacenter = inputs["metacentric_height"]
        Awater = inputs["platform_Awater"]
        K_moor = np.diag(inputs["mooring_stiffness"])

        # Compute elements on mass matrix diagonal
        # Surge, sway, heave just use normal inertia (without mooring according to Senu)
        # Roll, pitch, yaw use system moment of inertia
        M_mat = np.r_[m_system * np.ones(3), I_system[:3]]

        # Structural stiffness
        K_struct = np.zeros(nDOF)
        K_struct[3:5] = np.abs(m_system * gravity * z_cg)

        # Hydrostatic stiffness has contributions in heave (K33) and roll/pitch (K44/55)
        # See DNV-RP-H103: Modeling and Analyis of Marine Operations
        K_hydro = np.zeros(nDOF)
        K_hydro[2] = rhoWater * gravity * Awater
        K_hydro[3:5] = rhoWater * gravity * V_system * h_metacenter
        outputs["hydrostatic_stiffness"] = K_hydro

        # Now compute all six natural periods at once
        K_total = np.maximum(K_struct + K_hydro + K_moor, 0.0)
        outputs["rigid_body_periods"] = 2 * np.pi * np.sqrt((M_mat + A_mat) / (K_total + eps))
        outputs["surge_period"] = outputs["rigid_body_periods"][0]
        outputs["sway_period"] = outputs["rigid_body_periods"][1]
        outputs["heave_period"] = outputs["rigid_body_periods"][2]
        outputs["roll_period"] = outputs["rigid_body_periods"][3]
        outputs["pitch_period"] = outputs["rigid_body_periods"][4]
        outputs["yaw_period"] = outputs["rigid_body_periods"][5]
